import pandas as pd
import numpy as np
import math
import json
import datetime
import time
from functools import partial
from typing import Optional, List, Union, Dict
from multiprocessing import Pool
from .base import Factor
from .fund_basic_factors import *
from ...data.manager import DataManager
from ...data.struct import AssetTimeSpan
from ...data.manager.fund_info_filter import fund_info_update, active_fund_info as get_active_fund_info
from ...constant import FundFactorType 

class TradeYear(Factor):

    def __init__(self):
        super().__init__(f_name='TradeYear', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundNavDailyModify())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=1).apply(Calculator.rolling_trade_year, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FeeRate(Factor):

    def __init__(self):
        super().__init__(f_name='FeeRate', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfo())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundInfo().get()
        self._factor = self._factor[['fund_id','manage_fee','trustee_fee']].set_index('fund_id').fillna(0).sum(axis = 1)
        self._factor = pd.DataFrame(self._factor, columns=[datetime.date(2005,1,4)]).T
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AlphaBetaDaily3yI(Factor):
    # cost 20-30 mins
    def __init__(self):
        super().__init__(f_name='AlphaBetaDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfoAsset())
        self._deps.add(FundRetDailyModify())
        self._deps.add(IndexRetDaily())
        self._deps.add(TradingDay())
        year = 3
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def loop_item(self, fund_ret, index_ret, index_id, fund_id):
        df = fund_ret[[fund_id]].join(index_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'}).reset_index()
        res = []
        pd.Series(df.index).rolling(
            window=self.time_range, min_periods=Calculator.MIN_TIME_SPAN).apply(
            partial(Calculator.rolling_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{fund_id}', 'beta':f'beta_{fund_id}'})
        return df
    
    def loop_item_update(self, dt, fund_ret, index_ret, index_id, fund_id):
        df = fund_ret[[fund_id]].join(index_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'})
        df = df.loc[:dt].tail(self.time_range).reset_index()
        res = []
        Calculator._rolling_alpha_beta(res, df)
        fund_res = {f'alpha_{fund_id}':res[0]['alpha'],
                    f'beta_{fund_id}':res[0]['beta']}   
        return fund_res

    def update_one_day(self, dt, index_fund, fund_ret, index_ret):
        dic = {}
        res = []
        for index_id, fund_list in index_fund.items():  
            _res = [self.loop_item_update(dt, fund_ret, index_ret, index_id, i) for i in fund_list]
            res.extend(_res)
        for _i in res:
            dic.update(_i)
        print(f'\t\tfinish {dt}')
        return pd.DataFrame([dic], index=[dt])

    def append_update(self):
        self._factor = self.get()
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)

        last_date = self._factor.index[-1]
        td = fund_ret.index
        update_list = td[td > last_date].tolist()
        _df_list = []
        print('\tupdate_list  ',update_list)
        for dt in update_list:
            res = self.update_one_day(dt, index_fund, fund_ret, index_ret)
            _df_list.append(res)
        if _df_list == []:
            return True
        df_update = pd.concat(_df_list,axis=0)
        self._factor = self._factor.append(df_update)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        return self.save()

    def calc(self, is_multi_cpu=False):
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)
        
        # calculate
        result = []
        for index_id, fund_list in index_fund.items():
            if is_multi_cpu:
                p = Pool()
                res = [i for i in p.imap_unordered(partial(self.loop_item, fund_ret, index_ret, index_id), fund_list, 64)]
                p.close()
                p.join()
                result.extend(res)
            else:
                for fund_id in fund_list:
                    res_i = self.loop_item(fund_ret, index_ret, index_id, fund_id)
                    result.append(res_i)
        self._factor = pd.concat(result, axis=1)
        self._factor = self._factor.sort_index()


class AlphaDaily3yI(Factor):

    def __init__(self):
        super().__init__(f_name='AlphaDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AlphaBetaDaily3yI())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = AlphaBetaDaily3yI().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class BetaDaily3yI(Factor):

    def __init__(self):
        super().__init__(f_name='BetaDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AlphaBetaDaily3yI())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = AlphaBetaDaily3yI().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AlphaMdd3yI(Factor):

    def __init__(self):
        super().__init__(f_name='AlphaMdd3yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AlphaDaily3yI())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = AlphaDaily3yI().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_mdd, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class WinRateMonthlyI(Factor):
        
    def __init__(self):
        super().__init__(f_name='WinRateMonthlyI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfo())
        self._deps.add(ActiveFundInfo())
        self._deps.add(FundRetDailyModify())
        self._deps.add(IndexRetDaily())
        self._deps.add(FundInfoAsset())

    def calc(self):
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)
        
        fund_ret_monthly = Calculator.data_resample_monthly_ret(df=fund_ret, min_count=5).dropna(axis=0,how='all')
        index_ret_monthly = Calculator.data_resample_monthly_ret(df=index_ret, min_count=5).dropna(axis=0,how='all')
        res = []
        for index_id, fund_list in index_fund.items():
            con_df = ~ fund_ret_monthly[fund_list].isnull() * 1
            con_df = con_df.replace(0,np.nan)
            winrate = fund_ret_monthly[fund_list].gt(index_ret_monthly[index_id],axis='index') * 1
            res_i = (con_df * winrate).rolling(window=winrate.shape[0],min_periods=6).mean()
            res.append(res_i)
        self._factor = pd.concat(res, axis=1)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class WinRateMonthlyTop50(Factor):

    def __init__(self):
        super().__init__(f_name='WinRateMonthlyTop50', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfo())
        self._deps.add(ActiveFundInfo())
        self._deps.add(FundRetDailyModify())
        self._deps.add(IndexRetDaily())
        self._deps.add(FundInfoAsset())
        self._deps.add(TradingDay())
        self.top_pct = 0.5

    def calc(self):
        fund_ret = FundRetDailyModify().get()
        fund_ret_monthly = Calculator.data_resample_monthly_ret(fund_ret).dropna(axis=0,how='all')
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)
        
        res = []
        for index_id, fund_list in index_fund.items():
            _df = fund_ret_monthly[fund_list]
            con_df = ~ _df.isnull() * 1
            con_df = con_df.replace(0,np.nan)
            _df_rank = _df.rank(pct=True,axis=1).copy()
            top_df = (_df_rank > self.top_pct) * 1
            res_i = (con_df * top_df).rolling(window=top_df.shape[0],min_periods=6).mean()
            res.append(res_i)
        self._factor = pd.concat(res, axis=1)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class WinRateMonthlyTop75(Factor):

    def __init__(self):
        super().__init__(f_name='WinRateMonthlyTop75', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfo())
        self._deps.add(ActiveFundInfo())
        self._deps.add(FundRetDailyModify())
        self._deps.add(IndexRetDaily())
        self._deps.add(FundInfoAsset())
        self._deps.add(TradingDay())
        self.top_pct = 0.75

    def calc(self):
        fund_ret = FundRetDailyModify().get()
        fund_ret_monthly = Calculator.data_resample_monthly_ret(fund_ret).dropna(axis=0,how='all')
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)
        
        res = []
        for index_id, fund_list in index_fund.items():
            _df = fund_ret_monthly[fund_list]
            con_df = ~ _df.isnull() * 1
            con_df = con_df.replace(0,np.nan)
            _df_rank = _df.rank(pct=True,axis=1).copy()
            top_df = (_df_rank > self.top_pct) * 1
            res_i = (con_df * top_df).rolling(window=top_df.shape[0],min_periods=6).mean()
            res.append(res_i)
        self._factor = pd.concat(res, axis=1)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class TrackerrDaily3yI(Factor):

    def __init__(self):
        super().__init__(f_name='TrackerrDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfo())
        self._deps.add(ActiveFundInfo())
        self._deps.add(FundRetDailyModify())
        self._deps.add(IndexRetDaily())
        self._deps.add(TradingDay())
        self._deps.add(FundInfoAsset())
        year = 3
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def loop_item(self, fund_ret, index_ret, index_id, fund_id):
        df = fund_ret[[fund_id]].join(index_ret[[index_id]]).dropna()
        df['track_err'] = (df[fund_id] - df[index_id]).rolling(window=self.time_range, min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1)* np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        df = df[['track_err']].rename(columns={'track_err': fund_id})
        return df

    def loop_item_update(self, fund_ret, index_ret, index_id, dt, fund_id):
        df = fund_ret[[fund_id]].join(index_ret[[index_id]]).dropna()
        track_err = (df[fund_id] - df[index_id]).loc[:dt].tail(self.time_range).std(ddof=1)* np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        return {'fund_id':fund_id, 'track_err':track_err}

    def update_one_day(self, dt, index_fund, fund_ret, index_ret):
        dic = {}
        res = []
        for index_id, fund_list in index_fund.items():
            _res = [self.loop_item_update(fund_ret, index_ret, index_id, dt, i) for i in fund_list]
            res.extend(_res)
        _df = pd.DataFrame(res).set_index('fund_id').T
        _df.index = [dt]
        print(f'\t\tfinish {dt}')
        return _df
    
    def append_update(self):
        self._factor = self.get()
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)

        last_date = self._factor.index[-1]
        td = fund_ret.index
        update_list = td[td > last_date].tolist()
        print('\tupdate_list  ',update_list)
        _df_list = []
        for dt in update_list:
            res = self.update_one_day(dt, index_fund, fund_ret, index_ret)
            _df_list.append(res)
        if _df_list == []:
            return True
        df_update = pd.concat(_df_list,axis=0)
        self._factor = self._factor.append(df_update)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        return self.save()

    def calc(self, is_multi_cpu=False):
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)
  
        # calculate
        result = []
        for index_id, fund_list in index_fund.items():
            if is_multi_cpu:
                p = Pool()
                res = [i for i in p.imap_unordered(partial(self.loop_item, fund_ret, index_ret, index_id), fund_list, 64)]
                p.close()
                p.join()
                result.extend(res)
            else:
                for fund_id in fund_list:
                    res_i = self.loop_item(fund_ret, index_ret, index_id, fund_id)
                    result.append(res_i)
        self._factor = pd.concat(result, axis=1)
        self._factor = self._factor.sort_index()
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        
class InfoRatioDaily3yI(Factor):

    def __init__(self):
        super().__init__(f_name='InfoRatioDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AlphaDaily3yI())
        self._deps.add(TrackerrDaily3yI())
        self._deps.add(TradingDay())

    def calc(self):
        alpha = AlphaDaily3yI().get()
        track_err = TrackerrDaily3yI().get()
        common_funds = alpha.columns.intersection(track_err.columns)
        self._factor = alpha[common_funds] / track_err[common_funds]
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MddDaily3y(Factor):

    def __init__(self):
        super().__init__(f_name='MddDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundNavDailyModify())
        self._deps.add(TradingDay())
        year = 3
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_mdd, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class TreynorDaily3yI(Factor):

    def __init__(self):
        super().__init__(f_name='TreynorDaily3yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(BetaDaily3yI())
        self._deps.add(FundRetDailyModify())
        year = 3
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        beta = BetaDaily3yI().get()
        fund_ret = FundRetDailyModify().get()
        common_funds = beta.columns.intersection(fund_ret.columns)
        fund_mean_ret = fund_ret.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).mean()
        fund_mean_ret_no_free_ret = fund_mean_ret - Calculator.RISK_FEE_RATE_PER_DAY
        self._factor = fund_mean_ret_no_free_ret[common_funds] * Calculator.TRADING_DAYS_PER_YEAR / beta[common_funds]
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class DownriskDaily3y(Factor):

    def __init__(self):
        super().__init__(f_name='DownriskDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundRetDailyModify())
        year = 3
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor = np.abs(np.minimum(self._factor - Calculator.RISK_FEE_RATE_PER_DAY, 0))
        self._factor = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).mean()* np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class RetOverPeriodDaily3y(Factor):
    
    def __init__(self):
        super().__init__(f_name='RetOverPeriodDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundNavDailyModify())
        year = 3
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor / self._factor.fillna(method='bfill').shift(self.time_range) - 1
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AnnualRetDaily3y(Factor):

    def __init__(self):
        super().__init__(f_name='AnnualRetDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundNavDailyModify())
        year = 3
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_annual_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AnnualVolDaily3y(Factor):

    def __init__(self):
        super().__init__(f_name='AnnualVolDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundRetDailyModify())
        year = 3
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class SharpeDaily3y(Factor):

    def __init__(self):
        super().__init__(f_name='SharpeDaily3y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AnnualRetDaily3y())
        self._deps.add(AnnualVolDaily3y())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor =  (AnnualRetDaily3y().get() - Calculator.RISK_FEE_RATE) / AnnualVolDaily3y().get()
        self._factor = self._factor.replace(-np.Inf,np.nan).replace(np.Inf,np.nan)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AlphaBetaDaily1yI(Factor):

    def __init__(self):
        super().__init__(f_name='AlphaBetaDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfoAsset())
        self._deps.add(FundRetDailyModify())
        self._deps.add(IndexRetDaily())
        self._deps.add(TradingDay())
        year = 1
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def loop_item(self, fund_ret, index_ret, index_id, fund_id):
        df = fund_ret[[fund_id]].join(index_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'}).reset_index()
        res = []
        pd.Series(df.index).rolling(
            window=self.time_range, min_periods=Calculator.MIN_TIME_SPAN).apply(
            partial(Calculator.rolling_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{fund_id}', 'beta':f'beta_{fund_id}'})
        return df
    
    def loop_item_update(self, dt, fund_ret, index_ret, index_id, fund_id):
        df = fund_ret[[fund_id]].join(index_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'})
        df = df.loc[:dt].tail(self.time_range).reset_index()
        res = []
        Calculator._rolling_alpha_beta(res, df)
        fund_res = {f'alpha_{fund_id}':res[0]['alpha'],
                    f'beta_{fund_id}':res[0]['beta']}   
        return fund_res

    def update_one_day(self, dt, index_fund, fund_ret, index_ret):
        dic = {}
        res = []
        for index_id, fund_list in index_fund.items():  
            _res = [self.loop_item_update(dt, fund_ret, index_ret, index_id, i) for i in fund_list]
            res.extend(_res)
        for _i in res:
            dic.update(_i)
        print(f'\t\tfinish {dt}')
        return pd.DataFrame([dic], index=[dt])

    def append_update(self):
        self._factor = self.get()
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)

        last_date = self._factor.index[-1]
        td = fund_ret.index
        update_list = td[td > last_date].tolist()
        _df_list = []
        print('\tupdate_list  ',update_list)
        for dt in update_list:
            res = self.update_one_day(dt, index_fund, fund_ret, index_ret)
            _df_list.append(res)
        if _df_list == []:
            return True
        df_update = pd.concat(_df_list,axis=0)
        self._factor = self._factor.append(df_update)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        return self.save()

    def calc(self, is_multi_cpu=False):
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)
        
        # calculate
        result = []
        for index_id, fund_list in index_fund.items():
            if is_multi_cpu:
                p = Pool()
                res = [i for i in p.imap_unordered(partial(self.loop_item, fund_ret, index_ret, index_id), fund_list, 64)]
                p.close()
                p.join()
                result.extend(res)
            else:
                for fund_id in fund_list:
                    res_i = self.loop_item(fund_ret, index_ret, index_id, fund_id)
                    result.append(res_i)
        self._factor = pd.concat(result, axis=1)
        self._factor = self._factor.sort_index()
        
class AlphaDaily1yI(Factor):

    def __init__(self):
        super().__init__(f_name='AlphaDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AlphaBetaDaily1yI())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = AlphaBetaDaily1yI().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class BetaDaily1yI(Factor):

    def __init__(self):
        super().__init__(f_name='BetaDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AlphaBetaDaily1yI())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = AlphaBetaDaily1yI().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class TrackerrDaily1yI(Factor):

    def __init__(self):
        super().__init__(f_name='TrackerrDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfo())
        self._deps.add(ActiveFundInfo())
        self._deps.add(FundRetDailyModify())
        self._deps.add(IndexRetDaily())
        self._deps.add(TradingDay())
        self._deps.add(FundInfoAsset())
        year = 1
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def loop_item(self, fund_ret, index_ret, index_id, fund_id):
        df = fund_ret[[fund_id]].join(index_ret[[index_id]]).dropna()
        df['track_err'] = (df[fund_id] - df[index_id]).rolling(window=self.time_range, min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1)* np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        df = df[['track_err']].rename(columns={'track_err': fund_id})
        return df

    def loop_item_update(self, fund_ret, index_ret, index_id, dt, fund_id):
        df = fund_ret[[fund_id]].join(index_ret[[index_id]]).dropna()
        track_err = (df[fund_id] - df[index_id]).loc[:dt].tail(self.time_range).std(ddof=1)* np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        return {'fund_id':fund_id, 'track_err':track_err}

    def update_one_day(self, dt, index_fund, fund_ret, index_ret):
        dic = {}
        res = []
        for index_id, fund_list in index_fund.items():
            _res = [self.loop_item_update(fund_ret, index_ret, index_id, dt, i) for i in fund_list]
            res.extend(_res)
        _df = pd.DataFrame(res).set_index('fund_id').T
        _df.index = [dt]
        print(f'\t\tfinish {dt}')
        return _df
    
    def append_update(self):
        self._factor = self.get()
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)

        last_date = self._factor.index[-1]
        td = fund_ret.index
        update_list = td[td > last_date].tolist()
        print('\tupdate_list  ',update_list)
        _df_list = []
        for dt in update_list:
            res = self.update_one_day(dt, index_fund, fund_ret, index_ret)
            _df_list.append(res)
        if _df_list == []:
            return True
        df_update = pd.concat(_df_list,axis=0)
        self._factor = self._factor.append(df_update)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        return self.save()

    def calc(self, is_multi_cpu=False):
        fund_ret = FundRetDailyModify().get()
        index_ret = IndexRetDaily().get()
        fund_info_asset = FundInfoAsset().get()
        fund_info_asset = fund_info_asset[['fund_id','bk_index_id']].dropna()
        index_fund = {}
        for r in fund_info_asset.itertuples():
            if r.bk_index_id not in index_fund:
                index_fund[r.bk_index_id] = []
            index_fund[r.bk_index_id].append(r.fund_id)
  
        # calculate
        result = []
        for index_id, fund_list in index_fund.items():
            if is_multi_cpu:
                p = Pool()
                res = [i for i in p.imap_unordered(partial(self.loop_item, fund_ret, index_ret, index_id), fund_list, 64)]
                p.close()
                p.join()
                result.extend(res)
            else:
                for fund_id in fund_list:
                    res_i = self.loop_item(fund_ret, index_ret, index_id, fund_id)
                    result.append(res_i)
        self._factor = pd.concat(result, axis=1)
        self._factor = self._factor.sort_index()
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class InfoRatioDaily1yI(Factor):

    def __init__(self):
        super().__init__(f_name='InfoRatioDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AlphaDaily1yI())
        self._deps.add(TrackerrDaily1yI())
        self._deps.add(TradingDay())

    def calc(self):
        alpha = AlphaDaily1yI().get()
        track_err = TrackerrDaily1yI().get()
        common_funds = alpha.columns.intersection(track_err.columns)
        self._factor = alpha[common_funds] / track_err[common_funds]
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MddDaily1y(Factor):

    def __init__(self):
        super().__init__(f_name='MddDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundNavDailyModify())
        year = 1
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_mdd, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class TreynorDaily1yI(Factor):

    def __init__(self):
        super().__init__(f_name='TreynorDaily1yI', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundRetDailyModify())
        self._deps.add(BetaDaily3yI())
        year = 1
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        beta = BetaDaily3yI().get()
        self._factor = FundRetDailyModify().get()
        common_funds = beta.columns.intersection(self._factor.columns)
        fund_mean_ret = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).mean()
        fund_mean_ret_no_free_ret = fund_mean_ret - Calculator.RISK_FEE_RATE_PER_DAY
        self._factor = fund_mean_ret_no_free_ret[common_funds] * Calculator.TRADING_DAYS_PER_YEAR / beta[common_funds]
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class DownriskDaily1y(Factor):

    def __init__(self):
        super().__init__(f_name='DownriskDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundRetDailyModify())
        year = 1
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor = np.abs(np.minimum(self._factor - Calculator.RISK_FEE_RATE_PER_DAY, 0))
        self._factor = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).mean()* np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class RetOverPeriodDaily1y(Factor):
    
    def __init__(self):
        super().__init__(f_name='RetOverPeriodDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundNavDailyModify())
        year = 1
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor / self._factor.fillna(method='bfill').shift(self.time_range) - 1
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AnnualRetDaily1y(Factor):

    def __init__(self):
        super().__init__(f_name='AnnualRetDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundNavDailyModify())
        year = 1
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor =  FundNavDailyModify().get()
        self._factor = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_annual_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AnnualVolDaily1y(Factor):

    def __init__(self):
        super().__init__(f_name='AnnualVolDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundRetDailyModify())
        year = 1
        self.time_range = year * Calculator.TRADING_DAYS_PER_YEAR

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor = self._factor.rolling(window=self.time_range,min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class SharpeDaily1y(Factor):

    def __init__(self):
        super().__init__(f_name='SharpeDaily1y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AnnualRetDaily1y())
        self._deps.add(AnnualVolDaily1y())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor =  (AnnualRetDaily1y().get() - Calculator.RISK_FEE_RATE) /  AnnualVolDaily1y().get()
        self._factor = self._factor.replace(-np.Inf,np.nan).replace(np.Inf,np.nan)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AnnualRetDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='AnnualRetDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundNavDailyModify())

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_annual_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AnnualVolDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='AnnualVolDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundRetDailyModify())

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class AnnualRetMonthlyHalfY(Factor):
    # 不在日度做reindex 算ContinueRegValue用
    def __init__(self):
        super().__init__(f_name='AnnualRetMonthlyHalfY', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundNavMonthlyModify())

    def calc(self):
        self._factor = FundNavMonthlyModify().get()
        self._factor = self._factor.rolling(window=6).apply(Calculator.rolling_monthly_annual_ret, raw=True)
        self._factor = self._factor.dropna(axis=0, how='all')

class AnnualVolMonthlyHalfY(Factor):
    # 不在日度做reindex 算ContinueRegValue用
    def __init__(self):
        super().__init__(f_name='AnnualVolMonthlyHalfY', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundRetMonthlyModify())

    def calc(self):
        self._factor = FundRetMonthlyModify().get()
        self._factor = self._factor.rolling(window=6).std(ddof=1) * np.sqrt(12)
        self._factor = self._factor.dropna(axis=0, how='all')

class SharpeMonthlyHalfY(Factor):
    # 不在日度做reindex 算ContinueRegValue用
    def __init__(self):
        super().__init__(f_name='SharpeMonthlyHalfY', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(AnnualRetMonthlyHalfY())
        self._deps.add(AnnualVolMonthlyHalfY())

    def calc(self):
        self._factor =  (AnnualRetMonthlyHalfY().get() - Calculator.RISK_FEE_RATE) / AnnualVolMonthlyHalfY().get()
        self._factor = self._factor.replace(-np.Inf,np.nan).replace(np.Inf,np.nan)
        self._factor = self._factor.dropna(axis=0, how='all')

class ContinueRegValue(Factor):

    def __init__(self):
        super().__init__(f_name='ContinueRegValue', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(SharpeMonthlyHalfY())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = SharpeMonthlyHalfY().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=6).apply(Calculator.rolling_auto_reg, raw=True)
        self._factor = self._factor.dropna(axis=0, how='all')
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class TotalRetDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='TotalRetDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundNavDailyModify())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_totol_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MddDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MddDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundNavDailyModify())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundNavDailyModify().get() 
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_mdd, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class RecentMonthRet(Factor):
    
    def __init__(self):
        super().__init__(f_name='RecentMonthRet', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundNavDailyModify())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = self._factor.rolling(window=20,min_periods=5).apply(Calculator.rolling_totol_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class DownsideStdDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='DownsideStdDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundRetDailyModify())

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor[self._factor > 0] = 0
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())


class FundClAlphaBetaHistoryWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClAlphaBetaHistoryWeekly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfo())
        self._deps.add(FundRetWeeklyModify())
        self._deps.add(IndexRetWeekly())
        self._deps.add(TradingDay())
        self._deps.add(FundInfoAsset())

    def loop_item(self, fund_wind_class_dic, fund_weekly_ret, index_weekly_ret, fund_id):
        index_id = Calculator.WIND_TYPE_DICT[fund_wind_class_dic[fund_id]]
        df = fund_weekly_ret[[fund_id]].join(index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'}).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        pd.Series(df.index).rolling(
            window=df.shape[0], min_periods=24).apply(
            partial(Calculator.rolling_cl_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{fund_id}', 'beta':f'beta_{fund_id}'})
        return df

    def loop_item_update(self, fund_wind_class_dic, fund_weekly_ret, index_weekly_ret, dt, fund_id):
        index_id = Calculator.WIND_TYPE_DICT[fund_wind_class_dic[fund_id]]
        df = fund_weekly_ret[[fund_id]].join(index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'}).loc[:dt].reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        Calculator._rolling_cl_alpha_beta(res, df[['fund','benchmark']].to_numpy())
        fund_res = {f'alpha_{fund_id}':res[0]['alpha'],
                    f'beta_{fund_id}':res[0]['beta']}   
        return fund_res
    
    def update_one_day(self, dt, fund_weekly_ret, fund_wind_class_dic, index_weekly_ret):
        dic = {}
        fund_list = fund_weekly_ret.columns.tolist()
        result = [self.loop_item_update(fund_wind_class_dic, fund_weekly_ret, index_weekly_ret, dt, i) for i in fund_list]
        for _i in result:
            if _i is not None:
                dic.update(_i)
        print(f'\t\tfinish {dt}')
        return pd.DataFrame([dic], index=[dt])

    def append_update(self):
        self._factor = self.get()
        fund_info = FundInfo().get()
        fund_weekly_ret = FundRetWeeklyModify().get()
        index_weekly_ret = IndexRetWeekly().get().rename(columns=Calculator.INDEX_DICT)[list(Calculator.INDEX_DICT.values())]
        fund_wind_class_dic = fund_info.set_index('fund_id').to_dict()['wind_class_2']
        
        last_date = self._factor.index[-1]
        td = fund_weekly_ret.index
        update_list = td[td > last_date].tolist()
        _df_list = []
        print('\tupdate_list  ',update_list)
        for dt in update_list:
            res = self.update_one_day(dt, fund_weekly_ret, fund_wind_class_dic, index_weekly_ret)
            _df_list.append(res)
        if _df_list == []:
            return True
        df_update = pd.concat(_df_list,axis=0)
        self._factor = self._factor.append(df_update)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        return self.save()

    def calc(self, is_multi_cpu=False):
        fund_info = FundInfo().get()
        fund_weekly_ret = FundRetWeeklyModify().get()
        index_weekly_ret = IndexRetWeekly().get().rename(columns=Calculator.INDEX_DICT)[list(Calculator.INDEX_DICT.values())]
        fund_wind_class_dic = fund_info.set_index('fund_id').to_dict()['wind_class_2']
        
        fund_list = fund_weekly_ret.columns.tolist()
        # calculate
        result = []
        if is_multi_cpu:
            p = Pool()
            res = [i for i in p.imap_unordered(partial(self.loop_item, fund_wind_class_dic, fund_weekly_ret, index_weekly_ret), fund_list, 64) if i is not None]
            p.close()
            p.join()
            result.extend(res)
        else:
            for fund_id in fund_list:
                res_i = self.loop_item(fund_wind_class_dic, fund_weekly_ret, index_weekly_ret, fund_id)
                if res_i is not None:
                    result.append(res_i)
        self._factor = pd.concat(result, axis=1)
        self._factor = self._factor.sort_index()
        self._factor.index.name = 'datetime'
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundClAlphaHistoryWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClAlphaHistoryWeekly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundClAlphaBetaHistoryWeekly())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundClAlphaBetaHistoryWeekly().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundClBetaHistoryWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClBetaHistoryWeekly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundClAlphaBetaHistoryWeekly())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundClAlphaBetaHistoryWeekly().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundClAlphaBeta1YWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClAlphaBeta1YWeekly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundInfo())
        self._deps.add(FundRetWeeklyModify())
        self._deps.add(IndexRetWeekly())
        self._deps.add(TradingDay())

    def loop_item(self, fund_wind_class_dic, fund_weekly_ret, index_weekly_ret, fund_id):
        index_id = Calculator.WIND_TYPE_DICT[fund_wind_class_dic[fund_id]]
        df = fund_weekly_ret[[fund_id]].join(index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'}).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        pd.Series(df.index).rolling(
            window=52, min_periods=24).apply(
            partial(Calculator.rolling_cl_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{fund_id}', 'beta':f'beta_{fund_id}'})
        return df
    
    def loop_item_update(self, fund_wind_class_dic, fund_weekly_ret, index_weekly_ret, dt, fund_id):
        index_id = Calculator.WIND_TYPE_DICT[fund_wind_class_dic[fund_id]]
        df = fund_weekly_ret[[fund_id]].join(index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={fund_id:'fund',index_id:'benchmark'}).loc[:dt].reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        Calculator._rolling_cl_alpha_beta(res, df[['fund','benchmark']].to_numpy())
        fund_res = {f'alpha_{fund_id}':res[0]['alpha'],
                    f'beta_{fund_id}':res[0]['beta']}   
        return fund_res
    
    def update_one_day(self, dt, fund_weekly_ret, fund_wind_class_dic, index_weekly_ret):
        dic = {}
        fund_list = fund_weekly_ret.columns.tolist()
        result = [self.loop_item_update(fund_wind_class_dic, fund_weekly_ret, index_weekly_ret, dt, i) for i in fund_list]
        for _i in result:
            if _i is not None:
                dic.update(_i)
        print(f'\t\tfinish {dt}')
        return pd.DataFrame([dic], index=[dt])

    def append_update(self):
        self._factor = self.get()
        fund_info = FundInfo().get()
        fund_weekly_ret = FundRetWeeklyModify().get()
        index_weekly_ret = IndexRetWeekly().get().rename(columns=Calculator.INDEX_DICT)[list(Calculator.INDEX_DICT.values())]
        fund_wind_class_dic = fund_info.set_index('fund_id').to_dict()['wind_class_2']
        
        last_date = self._factor.index[-1]
        td = fund_weekly_ret.index
        update_list = td[td > last_date].tolist()
        _df_list = []
        print('\tupdate_list  ',update_list)
        for dt in update_list:
            res = self.update_one_day(dt, fund_weekly_ret, fund_wind_class_dic, index_weekly_ret)
            _df_list.append(res)
        if _df_list == []:
            return True
        df_update = pd.concat(_df_list,axis=0)
        self._factor = self._factor.append(df_update)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        return self.save()

    def calc(self, is_multi_cpu=False):
        fund_info = FundInfo().get()
        fund_weekly_ret = FundRetWeeklyModify().get()
        index_weekly_ret = IndexRetWeekly().get().rename(columns=Calculator.INDEX_DICT)[list(Calculator.INDEX_DICT.values())]
        fund_wind_class_dic = fund_info.set_index('fund_id').to_dict()['wind_class_2']
        
        fund_list = fund_weekly_ret.columns.tolist()
        # calculate
        result = []
        if is_multi_cpu:
            p = Pool()
            res = [i for i in p.imap_unordered(partial(self.loop_item, fund_wind_class_dic, fund_weekly_ret, index_weekly_ret), fund_list, 64) if i is not None]
            p.close()
            p.join()
            result.extend(res)
        else:
            for fund_id in fund_list:
                res_i = self.loop_item(fund_wind_class_dic, fund_weekly_ret, index_weekly_ret, fund_id)
                if res_i is not None:
                    result.append(res_i)
        self._factor = pd.concat(result, axis=1)
        self._factor = self._factor.sort_index()
        self._factor.index.name = 'datetime'
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundClAlpha1YWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClAlpha1YWeekly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundClAlphaBeta1YWeekly())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundClAlphaBeta1YWeekly().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundClBeta1YWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='FundClBeta1YWeekly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(FundClAlphaBetaHistoryWeekly())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = FundClAlphaBetaHistoryWeekly().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.split('_')[1] for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class ManagerIndex(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndex', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(IndexCloseDaily())
        self._deps.add(FundInfo())
        self._deps.add(FundManagerInfo())
        self._deps.add(FundSizeCombine())
        self._deps.add(FundNavDailyModify())
        self._deps.add(FundRetDailyModify())

    def fund_manager_prepare(self, start_date=None):
        mng_index_list: Dict[str, pd.DataFrame] = {}  # 基金经理指数
        mng_best_fund: Dict[str, pd.DataFrame] = {}  # 基金经理代表作 
        trading_days_list = TradingDay().get().index
        full_trading_days_list = DataManager.basic_data(func_name='get_trading_day_list').datetime
        full_trading_days_list = full_trading_days_list[full_trading_days_list >= datetime.date(2004,12,1)]
        index_price = IndexCloseDaily().get().rename(columns=Calculator.INDEX_DICT)
        index_ret = np.log(index_price).diff()
        fund_info = FundInfo().get()
        fund_manager_info = FundManagerInfo().get()
        fund_manager_info = fund_manager_info.rename(columns={'mng_id': 'manager_id', 'mng_name': 'name', 'start_date': 'start', 'end_date': 'end'})
        fund_info = fund_info[(fund_info.structure_type <= 1)
                                        & (~fund_info.wind_class_2.isnull())]
        fund_size = FundSizeCombine().get()
        fund_nav = FundNavDailyModify().get()
        fund_ret = FundRetDailyModify().get()
        if start_date is not None:
            index_price = index_price.loc[start_date:]
            index_ret = index_ret.loc[start_date:]
            fund_info = fund_info[fund_info.end_date >= start_date]
            fund_manager_info = fund_manager_info[fund_manager_info.end >= start_date]
            fund_size = fund_size.loc[start_date:]
            fund_nav = fund_nav.loc[start_date:]
            fund_ret = fund_ret.loc[start_date:]
            full_trading_days_list = full_trading_days_list[full_trading_days_list > (start_date - datetime.timedelta(days=5))]
        manager_name_dict = fund_manager_info.set_index('manager_id').to_dict()['name'] 
        fund_manager_info = fund_manager_info.set_index('fund_id')
        # 每个基金经理人数   如果多人给惩罚
        fund_list = fund_manager_info.index.unique().tolist()
        fund_manager_num_df = self.mng_process_manager_num(fund_list, fund_ret.index, fund_manager_info)
        
        # 每个基金的规模  小于1e 给惩罚
        fund_manager_size_df = fund_size.copy()
        fund_manager_size_df[fund_manager_size_df < 1e8] = 0.6
        fund_manager_size_df[(fund_manager_size_df >= 1e8) & (fund_manager_size_df < 2e9)] = 0.7
        fund_manager_size_df[(fund_manager_size_df >= 2e9) & (fund_manager_size_df < 5e9)] = 0.8
        fund_manager_size_df[(fund_manager_size_df >= 5e9) & (fund_manager_size_df < 1e10)] = 0.9
        fund_manager_size_df[fund_manager_size_df >= 1e10] = 1
        fund_manager_size_df = fund_manager_size_df.fillna(Calculator.PUNISH_RATIO)

        # 基金经理数据 根据被动主动分类
        pass_fund_id_list = fund_info[fund_info.wind_class_2.str.contains('被动')].fund_id.tolist()
        pass_fund_id_list = [i for i in pass_fund_id_list if i in fund_ret.columns] 
        acti_fund_id_list = [i for i in fund_ret.columns if i not in pass_fund_id_list] 
        _df_acti = pd.DataFrame(1, columns=acti_fund_id_list, index=fund_ret.index)
        _df_pass = pd.DataFrame(Calculator.PUNISH_RATIO, columns=pass_fund_id_list, index=fund_ret.index)
        fund_manager_stype_df = pd.concat([_df_acti, _df_pass], axis=1)
        
        for fund_type in Calculator.FUND_CLASSIFIER:
            mng_index_list[fund_type] = pd.DataFrame(index=fund_ret.index)
            mng_best_fund[fund_type] = pd.DataFrame(index=fund_ret.index)

        return fund_ret, fund_manager_size_df, fund_manager_num_df, fund_manager_stype_df, index_ret, full_trading_days_list, mng_best_fund, mng_index_list, manager_name_dict, fund_manager_info

    def mng_lambda_filter_date_range(self, single_fund_info: pd.DataFrame, date_index: pd.Series) -> pd.Series:
        result: List[pd.Series] = []
        for row in single_fund_info.itertuples(index=False):
            result.append(pd.Series((date_index >= row.start) & (date_index <= row.end), index=date_index))
        return pd.concat(result, axis=1).sum(axis=1)

    def mng_process_fund_ret(self, this_manager_fund_ids, this_manager_info, date_index):
        res = []
        for fund_id in this_manager_fund_ids:
            single_fund_info = this_manager_info.loc[[fund_id]]
            res_i =  self.mng_lambda_filter_date_range(single_fund_info, date_index)
            res_i.name = fund_id
            res.append(res_i)
        return pd.concat(res, axis=1)

    def mng_process_manager_num(self, fund_id_list, date_index, fund_manager_info):
        res = []
        for fund_id in fund_id_list:
            fund_managers_info = fund_manager_info.loc[[fund_id]]
            fund_manager_list = fund_managers_info.manager_id.unique()
            _res = []
            for _mng_id in fund_manager_list:
                fund_manager_i = fund_managers_info[fund_managers_info['manager_id'] == _mng_id]
                single_fund_res = self.mng_lambda_filter_date_range(fund_manager_i, date_index)
                single_fund_res = pd.DataFrame(single_fund_res, columns=[_mng_id])
                _res.append(single_fund_res)
            manager_in_same_time = pd.concat(_res,axis=1)
            fund_manger_same_time = pd.DataFrame(manager_in_same_time.sum(axis=1), columns=[fund_id])            
            res.append(fund_manger_same_time)
        fund_manger_same_time = pd.concat(res, axis=1)
        fund_manger_same_time[fund_manger_same_time > 1] = Calculator.PUNISH_RATIO
        return fund_manger_same_time

    def mng_process_manager_experience(self, fund_id_list, this_manager_info, date_index):
        res = []
        for fund_id in fund_id_list:
            single_fund_info = this_manager_info.loc[[fund_id]]
            work_day_df_i = self.mng_lambda_filter_date_range(single_fund_info, date_index)
            work_day_df_i.name = fund_id
            res.append(work_day_df_i)
        work_day_df = pd.DataFrame(res).T
        work_day_df = work_day_df.cumsum()
        low_cons = work_day_df < Calculator.TRADING_DAYS_PER_YEAR
        mid_cons = (work_day_df >= Calculator.TRADING_DAYS_PER_YEAR) & (work_day_df < 3 * Calculator.TRADING_DAYS_PER_YEAR)
        hig_cons = work_day_df >= 3 * Calculator.TRADING_DAYS_PER_YEAR
        work_day_df[low_cons] = Calculator.HARD_PUNISH_RATIO
        work_day_df[mid_cons] = Calculator.PUNISH_RATIO
        work_day_df[hig_cons] = 1
        return work_day_df

    def calc(self):
        fund_ret, fund_manager_size_df, fund_manager_num_df, fund_manager_stype_df, index_ret, full_trading_days_list, mng_best_fund, mng_index_list, manager_name_dict, fund_manager_info = self.fund_manager_prepare()
        fund_manager_list = list(manager_name_dict.keys())
        t0 = time.time()
        for m_id in fund_manager_list:
            _t0 = time.time()
            for fund_type in Calculator.FUND_CLASSIFIER:
                idx, _best_fund = self.index_calculation(m_id, fund_type, None, fund_manager_info, fund_ret, fund_manager_size_df, fund_manager_num_df,
                                             fund_manager_stype_df, index_ret, full_trading_days_list)
                if idx is not None:
                    mng_index_list[fund_type].loc[:, m_id] = idx # 该资产类别下记录计算结果
                    mng_best_fund[fund_type].loc[:, m_id] = _best_fund

            _idx = fund_manager_list.index(m_id)
            _t1 = time.time()
            #print(f'm_id {m_id} {_idx} {len(fund_manager_list)} cost time {_t1 - _t0}')
        t1 = time.time()
        #print(f'total cost {t1 - t0}')
        for fund_type in Calculator.FUND_CLASSIFIER:
            mng_index_list[fund_type] = mng_index_list[fund_type].replace(0, np.nan).dropna(how='all')
    
        res = []
        for fund_type, mng_index in mng_index_list.items():
            df = mng_index.copy()
            df.columns = [fund_type + '_' + i for i in df.columns]
            res.append(df)
        mng_index = pd.concat(res,axis=1).dropna(axis=1,how='all')
          
        res = []
        for fund_type, mng_index in mng_best_fund.items():
            df = mng_index.copy()
            print(fund_type)
            df.columns = ['bf_' + fund_type + '_' + i for i in df.columns] 
            res.append(df)
        mng_best_fund = pd.concat(res,axis=1).dropna(axis=1,how='all')
        self.mng_index = mng_index
        self.mng_best_fund = mng_best_fund
       
    def append_update(self):
        _exsited_factor = self.get()
        mng_index_uri = self._get_s3_factor_uri
        mng_best_fund_uri = mng_index_uri.replace('ManagerIndex','ManagerBestFund')
        _exstied_mng_best_fund = pd.read_parquet(mng_best_fund_uri)
        
        update_dt = _exsited_factor.index[-1]
        fund_ret, fund_manager_size_df, fund_manager_num_df, fund_manager_stype_df, index_ret, full_trading_days_list, mng_best_fund, mng_index_list, manager_name_dict, fund_manager_info = self.fund_manager_prepare(update_dt)
        fund_manager_list = list(manager_name_dict.keys())
        
        t0 = time.time()
        for m_id in fund_manager_list:
            _t0 = time.time()
            for fund_type in Calculator.FUND_CLASSIFIER:
                idx, _best_fund = self.index_calculation(m_id, fund_type, update_dt, fund_manager_info, fund_ret, fund_manager_size_df, fund_manager_num_df,
                                                                          fund_manager_stype_df, index_ret, full_trading_days_list)
                if idx is not None:
                    mng_index_list[fund_type].loc[:, m_id] = idx # 该资产类别下记录计算结果
                    mng_best_fund[fund_type].loc[:, m_id] = _best_fund

            _idx = fund_manager_list.index(m_id)
            _t1 = time.time()
            #print(f'm_id {m_id} {_idx} {len(fund_manager_list)} cost time {_t1 - _t0}')
        t1 = time.time()
        print(f'total cost time {t1 - t0}')

        res_df = []
        for fund_type, mng_index in mng_index_list.items():
            df = mng_index.copy()
            df.columns = [f'{fund_type}_{i}' for i in df.columns]
            df = df / df.iloc[0]
            df_part = (_exsited_factor[df.columns.intersection(_exsited_factor.columns)].iloc[-1] * df).iloc[1:]
            res_df.append(df_part)
        update_result = pd.concat(res_df,axis=1)
        mng_index = _exsited_factor.append(update_result)

        mng_best_fund_uri = mng_index_uri.replace('ManagerIndex','ManagerBestFund')
        _exstied_mng_best_fund = pd.read_parquet(mng_best_fund_uri)
        _res_best_fund = []
        for fund_type, best_df in mng_best_fund.items():
            df = best_df.copy()
            df.columns = [f'bf_{fund_type}_{i}' for i in df.columns]
            _res_best_fund.append(df.iloc[1:])
        mng_best_fund = _exstied_mng_best_fund.append(pd.concat(_res_best_fund,axis=1))
        self.mng_index = mng_index
        self.mng_best_fund = mng_best_fund
       
        return self.save()

    def clear(self, recursive=True):
        if recursive:
            for _dep in self._deps:
                _dep.clear(recursive)
        self.mng_index = None
        self.mng_best_fund = None

    def save(self):
        try:
            mng_index_uri = self._get_s3_factor_uri
            self.mng_index.to_parquet(mng_index_uri, compression='gzip')
            print(f' upload to s3 success {mng_index_uri}')

            mng_best_fund_uri = mng_index_uri.replace('ManagerIndex','ManagerBestFund')
            self.mng_best_fund.to_parquet(mng_best_fund_uri, compression='gzip')
            print(f' upload to s3 success {mng_best_fund_uri}')
            return True
        except:
            return False

    def index_calculation(self, m_id: str, 
                                fund_type: str, 
                                update_date: None, 
                                fund_manager_info: pd.DataFrame,
                                fund_ret: pd.DataFrame,
                                fund_manager_size_df: pd.DataFrame,
                                fund_manager_num_df: pd.DataFrame,
                                fund_manager_stype_df: pd.DataFrame,
                                index_ret: pd.DataFrame,
                                full_trading_days_list:pd.Series):
        wind_class_2_list = Calculator.FUND_CLASSIFIER[fund_type]
        # 该基金经理管理基金信息
        this_manager_info = fund_manager_info[(fund_manager_info['wind_class_2'].isin(wind_class_2_list))
                            & (fund_manager_info['manager_id'] == m_id)]
        if this_manager_info.shape[0] == 0:
             return None, None 
        index_start_date = this_manager_info.start.min()
        if update_date is not None:
            index_start_date = max(update_date, index_start_date)
        index_end_date = this_manager_info.end.max()
        # 该基金经理管理的同类基金列表
        this_manager_fund_ids = this_manager_info.index.unique().tolist()
        _fund_id_list1 = fund_ret.columns.tolist()
        _fund_id_list2 = fund_manager_size_df.columns.tolist()
        this_manager_fund_ids = list(set(this_manager_fund_ids).intersection(_fund_id_list1).intersection(_fund_id_list2))
        if len(this_manager_fund_ids) == 0:
             return None, None
        fund_ret = fund_ret.loc[index_start_date:index_end_date, this_manager_fund_ids]
        # 非管理时间，基金收益赋值0
        manager_status = self.mng_process_fund_ret(this_manager_fund_ids, this_manager_info, fund_ret.index)
        fund_ret = fund_ret * manager_status
        # 基金经理管理的基金 人数状态 单人为 1 ，多人为0.8
        manager_num = fund_manager_num_df.loc[index_start_date:index_end_date, this_manager_fund_ids]
        # 基金经理管理的基金 规模状态  > 100e : 1, > 50e : 0.9, > 20e : 0.8, > 1e : 0.7, < 1e : 0.6
        manager_size = fund_manager_size_df.loc[index_start_date:index_end_date, this_manager_fund_ids]
        # 基金经理管理的基金 负责年限 大于3 1， 大于1 0.8， 小于1 0.6
        manager_year = self.mng_process_manager_experience(this_manager_fund_ids, this_manager_info, fund_ret.index)
        # 基金经理管理的基金 主动 1 , 被动 0.8
        manager_style = fund_manager_stype_df.loc[index_start_date:index_end_date, this_manager_fund_ids]
        # 权重累乘
        fund_weight = manager_num * manager_size * manager_year * manager_style * manager_status
        # 权重和归1
        fund_weight['sum'] = fund_weight.sum(axis=1).T
        fund_weight = fund_weight[this_manager_fund_ids].divide(fund_weight['sum'], axis=0)
        # 基金经理 收益
        total_ret = (fund_ret * fund_weight).sum(axis=1)
        total_ret.name = m_id
        total_ret = pd.DataFrame(total_ret)
        if total_ret.shape[0] == 0:
            return None, None
        # 基金经理 收益 用指数填充待业期
        index_date_list = manager_status.index[manager_status.sum(axis=1) < 1]
        total_ret.loc[index_date_list, m_id] = index_ret.loc[index_date_list, fund_type]
        # 收益初值赋0
        last_trading_day = full_trading_days_list[full_trading_days_list < total_ret.index.values[0]].values[-1]
        total_ret.loc[last_trading_day] = 0
        total_ret = total_ret.sort_index()
        # 做指数
        fund_manager_index = np.exp(total_ret.cumsum())
        # 基金经理代表作
        manager_best_fund = pd.DataFrame((fund_weight + manager_status.cumsum()/1e8).idxmax(axis=1), columns=[m_id])
        #mng_best_fund[fund_type].loc[:, m_id] = manager_best_fund
        return fund_manager_index, manager_best_fund

class ManagerBestFund(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerBestFund', f_type=FundFactorType.DERIVED, f_level='derived')

    def calc(self):
        pass
    
    def save(self):
        return True

class ManagerIndexRetDaily(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexRetDaily', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(ManagerIndex())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = np.log(self._factor).diff(1)
        self._factor = Calculator.data_reindex_daily_trading_day_not_fill(self._factor, TradingDay().get())

class ManagerIndexWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexWeekly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndex())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = Calculator.data_resample_weekly_nav(self._factor).dropna(axis=0, how='all')

class ManagerIndexRetWeekly(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexWeekly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndexWeekly())

    def calc(self):
        self._factor = ManagerIndexWeekly().get()
        self._factor = np.log(self._factor).diff(1)

class ManagerIndexMonthly(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexMonthly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndex())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = Calculator.data_resample_monthly_nav(self._factor).dropna(axis=0, how='all')

class ManagerIndexRetMonthly(Factor):

    def __init__(self):
        super().__init__(f_name='ManagerIndexRetMonthly', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndexMonthly())

    def calc(self):
        self._factor = ManagerIndexMonthly().get()
        self._factor = np.log(self._factor).diff(1)

class MngAnnualRetDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngAnnualRetDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndex())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_annual_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngAnnualRetDaily1Y(Factor):

    def __init__(self):
        super().__init__(f_name='MngAnnualRetDaily1Y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndex())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = self._factor.rolling(window=Calculator.TRADING_DAYS_PER_YEAR,min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_annual_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngAnnualRetDaily3Y(Factor):

    def __init__(self):
        super().__init__(f_name='MngAnnualRetDaily3Y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndex())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = self._factor.rolling(window=Calculator.TRADING_DAYS_PER_YEAR*3,min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_annual_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngTotalRetDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngTotalRetDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndex())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_totol_ret, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngMddDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngMddDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndex())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).apply(Calculator.rolling_mdd, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngAnnualVolDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngAnnualVolDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndexRetDaily())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndexRetDaily().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngDownsideStdDailyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngDownsideStdDailyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndexRetDaily())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndexRetDaily().get().copy()
        self._factor[self._factor > 0] = 0
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngDownsideStdMonthlyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngDownsideStdMonthlyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndexRetMonthly())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndexRetMonthly().get().copy()
        self._factor[self._factor > 0] = 0
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=Calculator.MIN_TIME_SPAN).std(ddof=1) * np.sqrt(Calculator.TRADING_DAYS_PER_YEAR)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngFundTypeTradingDays(Factor):

    def __init__(self):
        super().__init__(f_name='MngFundTypeTradingDays', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndex())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = ManagerIndex().get()
        self._factor = self._factor.rolling(window=self._factor.shape[0],min_periods=1).apply(Calculator.rolling_trade_year, raw=True)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngClAlphaBetaWeekly1Y(Factor):

    def __init__(self):
        super().__init__(f_name='MngClAlphaBetaWeekly1Y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndexRetWeekly())
        self._deps.add(TradingDay())
        self._deps.add(IndexRetWeekly())

    def loop_item(self, mng_weekly_ret, index_weekly_ret, mng_id):
        index_id = mng_id.split('_')[0]
        df = mng_weekly_ret[[mng_id]].join(index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={mng_id:'fund',index_id:'benchmark'}).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        pd.Series(df.index).rolling(
            window=52, min_periods=24).apply(
            partial(Calculator.rolling_cl_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{mng_id}', 'beta':f'beta_{mng_id}'})
        return df

    def loop_item_update(self, mng_weekly_ret, index_weekly_ret, dt, mng_id):
        index_id = mng_id.split('_')[0]
        df = mng_weekly_ret[[mng_id]].join(index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={mng_id:'fund',index_id:'benchmark'}).loc[:dt].tail(52).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        Calculator._rolling_cl_alpha_beta(res, df[['fund','benchmark']].to_numpy())
        fund_res = {f'alpha_{mng_id}':res[0]['alpha'],
                    f'beta_{mng_id}':res[0]['beta']}   
        return fund_res
    
    def update_one_day(self, dt, mng_weekly_ret, index_weekly_ret):
        mng_list = mng_weekly_ret.columns.tolist()
        res = [self.loop_item_update(mng_weekly_ret, index_weekly_ret, dt, i) for i in mng_list]
        dic = {}
        for _i in res:
            if _i is not None:
                dic.update(_i)
        print(f'\t\tfinish {dt}')
        return pd.DataFrame([dic], index=[dt])

    def append_update(self):
        self._factor  = self.get()
        mng_weekly_ret = ManagerIndexRetWeekly().get()
        index_weekly_ret = IndexRetWeekly().get().rename(columns=Calculator.INDEX_DICT)[list(Calculator.INDEX_DICT.values())]
        last_date = self._factor.index[-1]
        td = index_weekly_ret.index
        update_list = td[td > last_date].tolist()
        _df_list = []
        print('\tupdate_list  ',update_list)
        for dt in update_list:
            res = self.update_one_day(dt, mng_weekly_ret, index_weekly_ret)
            _df_list.append(res)
        if _df_list == []:
            return True
        df_update = pd.concat(_df_list,axis=0)
        self._factor  = self._factor .append(df_update)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        return self.save()

    def calc(self, is_multi_cpu=False):
        mng_weekly_ret = ManagerIndexRetWeekly().get()
        index_weekly_ret = IndexRetWeekly().get().rename(columns=Calculator.INDEX_DICT)[list(Calculator.INDEX_DICT.values())]
        mng_list = mng_weekly_ret.columns.tolist()
        # calculate
        result = []
        if is_multi_cpu:
            p = Pool()
            res = [i for i in p.imap_unordered(partial(self.loop_item, mng_weekly_ret, index_weekly_ret), mng_list, 64) if i is not None]
            p.close()
            p.join()
            result.extend(res)
        else:
            for mng_id in mng_list:
                res_i = self.loop_item(mng_weekly_ret, index_weekly_ret, mng_id)
                if res_i is not None:
                    result.append(res_i)
        self._factor = pd.concat(result, axis=1)
        self._factor = self._factor.sort_index()
        self._factor.index.name = 'datetime'
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngClAlphaWeekly1Y(Factor):

    def __init__(self):
        super().__init__(f_name='MngClAlphaWeekly1Y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(MngClAlphaBetaWeekly1Y())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = MngClAlphaBetaWeekly1Y().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.replace('alpha_','') for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngClBetaWeekly1Y(Factor):

    def __init__(self):
        super().__init__(f_name='MngClBetaWeekly1Y', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(MngClAlphaBetaWeekly1Y())
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = MngClAlphaBetaWeekly1Y().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.replace('beta_','') for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngClAlphaBetaWeeklyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngClAlphaBetaWeeklyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(ManagerIndexRetWeekly())
        self._deps.add(TradingDay())
        self._deps.add(IndexRetWeekly())

    def loop_item(self, mng_weekly_ret, index_weekly_ret, mng_id):
        index_id = mng_id.split('_')[0]
        df = mng_weekly_ret[[mng_id]].join(index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={mng_id:'fund',index_id:'benchmark'}).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        pd.Series(df.index).rolling(
            window=df.shape[0], min_periods=24).apply(
            partial(Calculator.rolling_cl_alpha_beta, res=res, df=df), raw=True)
        df = pd.DataFrame(res,index=df.datetime[-len(res):])
        df = df.rename(columns={'alpha':f'alpha_{mng_id}', 'beta':f'beta_{mng_id}'})
        return df

    def loop_item_update(self, mng_weekly_ret, index_weekly_ret, dt, mng_id):
        index_id = mng_id.split('_')[0]
        df = mng_weekly_ret[[mng_id]].join(index_weekly_ret[[index_id]]).dropna()
        df = df.rename(columns={mng_id:'fund',index_id:'benchmark'}).loc[:dt].tail(52).reset_index()
        if df.shape[0] < 24:
            return None
        res = []
        Calculator._rolling_cl_alpha_beta(res, df[['fund','benchmark']].to_numpy())
        fund_res = {f'alpha_{mng_id}':res[0]['alpha'],
                    f'beta_{mng_id}':res[0]['beta']}   
        return fund_res
    
    def update_one_day(self, dt, mng_weekly_ret, index_weekly_ret):
        mng_list = mng_weekly_ret.columns.tolist()
        res = [self.loop_item_update(mng_weekly_ret, index_weekly_ret, dt, i) for i in mng_list]
        dic = {}
        for _i in res:
            if _i is not None:
                dic.update(_i)
        print(f'\t\tfinish {dt}')
        return pd.DataFrame([dic], index=[dt])

    def append_update(self):
        self._factor  = self.get()
        mng_weekly_ret = ManagerIndexRetWeekly().get()
        index_weekly_ret = IndexRetWeekly().get().rename(columns=Calculator.INDEX_DICT)[list(Calculator.INDEX_DICT.values())]
        last_date = self._factor.index[-1]
        td = index_weekly_ret.index
        update_list = td[td > last_date].tolist()
        _df_list = []
        print('\tupdate_list  ',update_list)
        for dt in update_list:
            res = self.update_one_day(dt, mng_weekly_ret, index_weekly_ret)
            _df_list.append(res)
        if _df_list == []:
            return True
        df_update = pd.concat(_df_list,axis=0)
        self._factor  = self._factor .append(df_update)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        return self.save()

    def calc(self, is_multi_cpu=False):
        mng_weekly_ret = ManagerIndexRetWeekly().get()
        index_weekly_ret = IndexRetWeekly().get().rename(columns=Calculator.INDEX_DICT)[list(Calculator.INDEX_DICT.values())]
        mng_list = mng_weekly_ret.columns.tolist()
        # calculate
        result = []
        if is_multi_cpu:
            p = Pool()
            res = [i for i in p.imap_unordered(partial(self.loop_item, mng_weekly_ret, index_weekly_ret), mng_list, 64) if i is not None]
            p.close()
            p.join()
            result.extend(res)
        else:
            for mng_id in mng_list:
                res_i = self.loop_item(mng_weekly_ret, index_weekly_ret, mng_id)
                if res_i is not None:
                    result.append(res_i)
        self._factor = pd.concat(result, axis=1)
        self._factor = self._factor.sort_index()
        self._factor.index.name = 'datetime'
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngClAlphaWeeklyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngClAlphaWeeklyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(MngClAlphaBetaWeeklyHistory())

    def calc(self):
        self._factor = MngClAlphaBetaWeeklyHistory().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'alpha']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.replace('alpha_','') for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngClBetaWeeklyHistory(Factor):

    def __init__(self):
        super().__init__(f_name='MngClBetaWeeklyHistory', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(MngClAlphaBetaWeeklyHistory())

    def calc(self):
        self._factor = MngClAlphaBetaWeeklyHistory().get()
        col_name = self._factor.columns.tolist()
        factor_list = [i for i in col_name if i.split('_')[0] == 'beta']
        self._factor = self._factor[factor_list]
        name_dic = { i : i.replace('beta_','') for i in factor_list}
        self._factor = self._factor[factor_list].rename(columns=name_dic)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class MngFundSize(Factor):

    def __init__(self):
        super().__init__(f_name='MngFundSize', f_type=FundFactorType.DERIVED, f_level='derived')
        self._deps.add(TradingDay())
        self._deps.add(FundInfo())
        self._deps.add(FundManagerInfo())
        self._deps.add(FundSize())

    def mng_process_fund_ret(self, this_manager_fund_ids, this_manager_info, date_index):
        res = []
        for fund_id in this_manager_fund_ids:
            single_fund_info = this_manager_info.loc[[fund_id]]
            res_i =  self.mng_lambda_filter_date_range(single_fund_info, date_index)
            res_i.name = fund_id
            res.append(res_i)
        return pd.concat(res, axis=1)

    def fund_manager_size_prepare(self):
        trading_days_list = TradingDay().get().index
        fund_info = FundInfo().get()
        fund_manager_info = FundManagerInfo().get()
        fund_manager_info = fund_manager_info.rename(columns={'mng_id': 'manager_id', 'mng_name': 'name', 'start_date': 'start', 'end_date': 'end'})
        ## 计算基金经理指数时只去掉分级基金子基金， 不能回测交易的基金依然影响基金经理能力
        fund_info = fund_info[(fund_info.structure_type <= 1)
                                        & (~fund_info.wind_class_2.isnull())]
        fund_size = FundSize().get()
        fund_manager_info = fund_manager_info.set_index('fund_id')
        manager_name_dict = fund_manager_info.set_index('manager_id').to_dict()['name'] 
        return manager_name_dict, fund_manager_info, fund_size, fund_info


    def calc(self):
        manager_name_dict, fund_manager_info, fund_size, fund_info = self.fund_manager_size_prepare()
        fund_manager_list = list(manager_name_dict.keys())
        t0 = time.time()
        loop_paras = []
        for m_id in fund_manager_list:
            for fund_type in Calculator.FUND_CLASSIFIER:
                if fund_type == 'index':
                    continue
                loop_paras.append([m_id, fund_type])
        result = []
        for para_i in loop_paras:
            df = self.size_calculation(fund_manager_info, fund_size, para_i, fund_info) 
            result.append(df)
        t1 = time.time()
        print(f'cost time {t1 - t0}')
        self._factor = pd.concat(result,axis=1).dropna(axis=1,how='all')

    def mng_lambda_filter_date_range(self, single_fund_info: pd.DataFrame, date_index: pd.Series) -> pd.Series:
        result: List[pd.Series] = []
        for row in single_fund_info.itertuples(index=False):
            result.append(pd.Series((date_index >= row.start) & (date_index <= row.end), index=date_index))
        return pd.concat(result, axis=1).sum(axis=1)

    def size_calculation(self, fund_manager_info, fund_size, para_i, fund_info):
        fund_info = fund_info[fund_info.structure_type.isin([0,1])]
        fund_list = fund_info.fund_id.tolist()
        m_id = para_i[0]
        fund_type = para_i[1]
        wind_class_2_list = Calculator.FUND_CLASSIFIER[fund_type]
        this_manager_info = fund_manager_info[(fund_manager_info['wind_class_2'].isin(wind_class_2_list))
                            & (fund_manager_info['manager_id'] == m_id)]
        if this_manager_info.shape[0] == 0:
            return None
        index_start_date = this_manager_info.start.min()
        index_end_date = this_manager_info.end.max()
        this_manager_fund_ids = this_manager_info.index.unique().intersection(fund_size.columns).intersection(fund_list).tolist()
        if len(this_manager_fund_ids) < 1:
            return None
        fund_size = fund_size.loc[index_start_date:index_end_date, this_manager_fund_ids]
        manager_status = self.mng_process_fund_ret(this_manager_fund_ids, this_manager_info, fund_size.index)
        fund_size = fund_size * manager_status
        fund_size = pd.DataFrame(fund_size.sum(axis=1), columns=[f'{fund_type}_{m_id}'])
        return fund_size




from typing import Dict

import numpy as np
import pandas as pd

from .factor_types import FinQualityFactor, MixFactorBase
from .basic_factors import GrossProfitFactor, IncomeTTMFactor, NetProfitTTMFactor, AdjNetAssetAvgFactor, TotalAssetAvgFactor, FixAssetAvgFactor
from .utils import normalize


class GrossMarginFactor(FinQualityFactor):
    # 毛利率
    def __init__(self):
        super().__init__('gr_mar')
        self._deps.add(GrossProfitFactor())
        self._deps.add(IncomeTTMFactor())

    def calc(self):
        # 过去四个季度的毛利润除以营业收入
        self._factor = GrossProfitFactor().get() / IncomeTTMFactor().get().replace({0: np.nan})


class ReturnOnAssetsFactor(FinQualityFactor):
    # 资产回报率
    def __init__(self):
        super().__init__('roa')
        self._deps.add(NetProfitTTMFactor())
        self._deps.add(TotalAssetAvgFactor())

    def calc(self):
        # 过去四个季度的净利润除以总资产
        self._factor = NetProfitTTMFactor().get() / TotalAssetAvgFactor().get().replace({0: np.nan})


class ReturnOnEquityFactor(FinQualityFactor):
    # 股本回报率
    def __init__(self):
        super().__init__('roe')
        self._deps.add(NetProfitTTMFactor())
        self._deps.add(AdjNetAssetAvgFactor())

    def calc(self):
        # 过去四个季度的净利润除以调整净资产
        self._factor = NetProfitTTMFactor().get() / AdjNetAssetAvgFactor().get().replace({0: np.nan})


class TurnoverRateOfFAFactor(FinQualityFactor):
    # 固定资产周转率
    def __init__(self):
        super().__init__('tr_fa')
        self._deps.add(IncomeTTMFactor())
        self._deps.add(FixAssetAvgFactor())

    def calc(self):
        # 过去四个季度的营业收入除以固定资产
        self._factor = IncomeTTMFactor().get() / FixAssetAvgFactor().get().replace({0: np.nan})


class MixFinQualityFactor(MixFactorBase, FinQualityFactor):
    # 合成因子
    def __init__(self):
        super(MixFactorBase, self).__init__('mixf')
        self._factor: Dict[pd.DataFrame] = {}
        self._deps.add(GrossMarginFactor())
        self._deps.add(ReturnOnAssetsFactor())
        self._deps.add(ReturnOnEquityFactor())
        self._deps.add(TurnoverRateOfFAFactor())

    def calc(self, universe: str):
        index = GrossMarginFactor().get(universe).index
        columns = GrossMarginFactor().get(universe).columns
        self._factor[universe] = pd.DataFrame((normalize(GrossMarginFactor().get(universe).T) + normalize(ReturnOnAssetsFactor().get(universe).T) +
                                               normalize(ReturnOnEquityFactor().get(universe).T) + normalize(TurnoverRateOfFAFactor().get(universe).T)).T, index=index, columns=columns)



import numpy as np
import pandas as pd
from statsmodels.tsa.ar_model import AutoReg
import statsmodels.api as sm

class Calculator:

    TRADING_DAYS_PER_YEAR = 242
    MIN_TIME_SPAN = int(TRADING_DAYS_PER_YEAR / 4)#为了延长基金回测范围，评分最低年限3个月
    RISK_FEE_RATE = 0.025
    RISK_FEE_RATE_PER_DAY = RISK_FEE_RATE / TRADING_DAYS_PER_YEAR
    PUNISH_RATIO = 0.8
    HARD_PUNISH_RATIO = 0.6
    FUND_CLASSIFIER = {
        'stock': ['普通股票型基金', '偏股混合型基金',  '被动指数型基金', '增强指数型基金', '平衡混合型基金', '灵活配置型基金', '股票多空'],
        'bond': ['中长期纯债型基金', '短期纯债型基金', '偏债混合型基金', '增强指数型债券基金', '被动指数型债券基金','混合债券型二级基金', '混合债券型一级基金'], 
        'QDII': ['国际(QDII)股票型基金', '国际(QDII)债券型基金', '国际(QDII)另类投资基金', '国际(QDII)混合型基金', ],
        'mmf': ['货币市场型基金'],
        'index': ['被动指数型基金', '增强指数型基金', '增强指数型债券基金', '被动指数型债券基金', '商品型基金', 'REITs'],
    }

    FILTER_NAV_RATIO = {
        'stock' : 0.1,
        'QDII'  : 0.1,
        'index' : 0.1,
        'bond'  : 0.07,
        'mmf'   : 0.01,
    }

    WIND_TYPE_DICT = {}
    for type_i, type_list in FUND_CLASSIFIER.items():
        for _ in type_list:
            WIND_TYPE_DICT.update({_:type_i})

    REPLACE_DICT = {'stock': 'csi_stockfund',
                     'bond': 'csi_boodfund', 
                     'mmf':  'mmf', 
                     'QDII': 'csi_f_qdii', 
                     'index':'hs300'}

    INDEX_DICT = { v : k for k, v in REPLACE_DICT.items()}

    @staticmethod
    def rolling_alpha_beta(x, res, df):
        df_i = df.loc[x[0]:x[-1],]
        return Calculator._rolling_alpha_beta(res, df_i)

    @staticmethod
    def rolling_cl_alpha_beta(x, res, df):
        df_i = df.loc[x[0]:x[-1],][['fund','benchmark']].to_numpy()
        return Calculator._rolling_cl_alpha_beta(res, df_i)

    @staticmethod
    def rolling_tm_alpha_beta(x, res, df):
        df_i = df.loc[x[0]:x[-1],][['fund','benchmark']].to_numpy()
        return Calculator._rolling_tm_alpha_beta(res, df_i)

    @staticmethod
    def rolling_hm_alpha_beta(x, res, df):
        df_i = df.loc[x[0]:x[-1],][['fund','benchmark']].to_numpy()
        return Calculator._rolling_hm_alpha_beta(res, df_i)

    @staticmethod
    def rolling_mdd(x):
        # fmax nanmin ignore nan
        x_max = np.fmax.accumulate(x, axis=0)
        return 1 - np.nanmin(x / x_max)

    @staticmethod
    def rolling_annual_ret(x):
        not_null_x = x[~np.isnan(x)]
        year = not_null_x.shape[0] / Calculator.TRADING_DAYS_PER_YEAR
        last_p = x[-1]
        first_p = not_null_x[0]
        return np.exp(np.log(last_p/first_p)/year) - 1
    
    @staticmethod
    def rolling_monthly_annual_ret(x):
        not_null_x = x[~np.isnan(x)]
        year = not_null_x.shape[0] / 12
        last_p = x[-1]
        first_p = not_null_x[0]
        return np.exp(np.log(last_p/first_p)/year) - 1
    
    @staticmethod
    def rolling_totol_ret(x):
        not_null_x = x[~np.isnan(x)]
        last_p = x[-1]
        first_p = not_null_x[0]
        return last_p / first_p

    @staticmethod
    def rolling_trade_year(x):
        not_null_x = x[~np.isnan(x)]
        year = not_null_x.shape[0] / Calculator.TRADING_DAYS_PER_YEAR
        return year

    @staticmethod
    def rolling_auto_reg(x):
        not_null_x = x[~np.isnan(x)]
        if len(not_null_x) < 6:
            return np.nan
        mod = AutoReg(endog=not_null_x, lags=1)
        res = mod.fit()
        continue_regress_v = res.params[0]
        return continue_regress_v

    @staticmethod
    def _rolling_alpha_beta(res, df_i):
        if sum(df_i.fund) == 0:
            res.append({'alpha':np.Inf,'beta':np.Inf})
            return 1
        else:
            ploy_res = np.polyfit(y=df_i.fund, x=df_i.benchmark,deg=1)
            p = np.poly1d(ploy_res)
            beta = ploy_res[0]
            alpha = ploy_res[1] * Calculator.TRADING_DAYS_PER_YEAR
            res.append({'alpha':alpha,'beta':beta})
            return 1
    
    @staticmethod
    def _rolling_cl_alpha_beta(res, total: np.ndarray):
        if total.shape[0] <= 1:
            res.append({'beta':np.nan,
                    'alpha':np.nan,
                    })
            return 1
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 0][X[:, 0] < 0] = 0
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            res.append({'beta':np.nan,
                    'alpha':np.nan,
                    })
            return 1
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        res.append({'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                })
        return 1
    
    @staticmethod
    def _rolling_tm_alpha_beta(res, total: np.ndarray):
        if total.shape[0] <= 1:
            res.append({'beta':np.nan,
                    'alpha':np.nan,
                    })
            return 1
        X = np.array([total[:, 1], total[:, 1]]).T
        X = np.array([total[:, 1], total[:, 1] * total[:, 1]]).T
        if np.count_nonzero(X[:, 1]) == 0:
            res.append({'beta':np.nan,
                    'alpha':np.nan,
                    })
            return 1
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        res.append({'beta':est2.params[1],
                'alpha':est2.params[-1],
                })
        return 1

    @staticmethod
    def _rolling_hm_alpha_beta(res, total: np.ndarray):
        if total.shape[0] <= 1:
            res.append({'beta':np.nan,
                        'alpha':np.nan,
                        })
            return 1
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 1][X[:, 1] < 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            res.append({'beta':np.nan,
                        'alpha':np.nan,
                        })
            return 1
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        res.append({'beta':est2.params[1],
                    'alpha':est2.params[-1],
                    })
        return 1

    @staticmethod
    def data_reindex_daily_trading_day(data, tradingday):
        # 日线因子交易日对齐
        td = pd.to_datetime(data.index)
        td = [i.date() for i in td]
        data.index = td
        data = data.reindex(data.index.union(tradingday.index))
        data = data.ffill().reindex(tradingday.index)
        return data

    @staticmethod
    def data_reindex_daily_trading_day_not_fill(data, tradingday):
        td = pd.to_datetime(data.index)
        td = [i.date() for i in td]
        data.index = td
        data = data.reindex(data.index.union(tradingday.index))
        data = data.reindex(tradingday.index)
        return data

    @staticmethod
    def data_resample_monthly_ret(df, rule='1M', min_count=15):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_weekly_ret(df, rule='1W', min_count=3):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_monthly_nav(df, rule='1M'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    @staticmethod
    def data_resample_weekly_nav(df, rule='1W'):
        df = df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).last()
        df.index = [i.date() for i in df.index]
        df.index.name = 'datetime'
        return df

    



import datetime
from typing import List, Optional
import traceback

import numpy as np
import pandas as pd

from ....constant import IndClassType, SEMI_UPDATE_DATE_LIST
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...view.derived_models import FundIndustryExposure
from ..raw.raw_data_helper import RawDataHelper
from .derived_data_helper import DerivedDataHelper


class FundIndustryExposureProcess:

    def __init__(self, data_helper: DerivedDataHelper):
        self._data_helper = data_helper
        self._result: Optional[pd.DataFrame] = None

    def init(self, start_date: str, end_date: str):
        self._end_date: str = end_date

        # 计算过去离目前最近的数据公布日期
        real_date, aux_date = RawDataHelper.get_prev_and_aux_target_dates(end_date, SEMI_UPDATE_DATE_LIST)
        # 基金的股票持仓明细
        fund_stock_portfolio = BasicDataApi().get_fund_stock_portfolio(dt=real_date).drop(columns=['datetime', '_update_time'])
        fund_stock_portfolio_aux = BasicDataApi().get_fund_stock_portfolio(dt=aux_date).drop(columns=['datetime', '_update_time'])
        fund_stock_portfolio_aux = fund_stock_portfolio_aux[~fund_stock_portfolio_aux.fund_id.isin(set(fund_stock_portfolio.fund_id.array))]
        self._fund_stock_portfolio = pd.concat([fund_stock_portfolio, fund_stock_portfolio_aux])
        stock_list: List[str] = list(self._fund_stock_portfolio.stock_id.unique())
        # 获取涉及到的这些股票的信息
        stock_infos = RawDataApi().get_em_stock_info(stock_list=stock_list)
        # 目前只需要申万一级行业分类ID
        sws_ind_code = stock_infos.set_index('stock_id')['bl_sws_ind_code']
        self._sws_ind_code = sws_ind_code.transform(lambda x: x.split('-')[0] if not pd.isnull(x) else np.nan)
        # 获取股票未复权价格
        stock_price = RawDataApi().get_em_stock_price(start_date=(pd.to_datetime(start_date).date() - datetime.timedelta(days=30)).isoformat(), end_date=end_date, stock_list=stock_list, columns=['close'])
        stock_price = stock_price.pivot(index='datetime', columns='stock_id', values='close').sort_index().ffill().iloc[[-1], :]
        self._stock_prices = stock_price.stack().droplevel('datetime').rename('close')

    def calc(self):
        def _aggregate_data(x):
            # valid_code = self._sws_ind_code.index.intersection(x.index).intersection(self._stock_prices.index)
            # return x[x.index.isin(valid_code)].loc[:, ['hold_number']].join(self._stock_prices.loc[valid_code]).join(self._sws_ind_code.loc[valid_code])
            return x.set_index('stock_id').loc[:, ['hold_number']].join([self._stock_prices, self._sws_ind_code], how='left')

        def _calc_ratio(x):
            x['value'] = x.mv / x.mv.sum()
            return x.value.droplevel('fund_id').to_json()

        # 把计算所需数据聚合在一起
        agg_result: pd.DataFrame = self._fund_stock_portfolio.groupby(by='fund_id', sort=False).apply(_aggregate_data)
        # 计算各持仓最新市值
        before_next_calc = agg_result.reset_index().drop(columns='stock_id')
        before_next_calc['mv'] = before_next_calc.hold_number * before_next_calc.close
        before_next_calc = before_next_calc.drop(columns=['hold_number', 'close'])
        # 计算每只基金各行业市值总和
        industry_sum = before_next_calc.groupby(by=['fund_id', 'bl_sws_ind_code'], sort=False).sum(min_count=1)
        # 计算每只基金各行业持仓比例
        self._result = industry_sum.groupby(level='fund_id', sort=False).apply(_calc_ratio).to_frame('value')
        # 这里先直接填成SWI
        self._result['ind_class_type'] = IndClassType.SWI1
        self._result['datetime'] = self._end_date

    def process(self, start_date: str, end_date: str, end_date_dt: datetime.date) -> List[str]:
        print(f'fund industry exposure update on the last day of week {end_date_dt}')
        failed_tasks: List[str] = []
        try:
            self.init(start_date, end_date)
            self.calc()
            self._data_helper._upload_derived(self._result.reset_index(), FundIndustryExposure.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append(f'fund_industry_exposure')
        return failed_tasks


if __name__ == '__main__':
    date = '20201231'
    fsg = FundIndustryExposureProcess(DerivedDataHelper())
    fsg.process(date, date, pd.to_datetime(date).date())



import numpy as np
import pandas as pd

from .factor_types import Alpha101Factor
from .basic_factors import StockPostPriceFactor, StockHighPriceFactor, StockOpenPriceFactor, StockLowPriceFactor, StockVolumeFactor, StockVWAPFactor, MarketValueFactor
from .utils import normalize


def Brank_cs(x):
    return (x.rank(axis=1).T/x.rank(axis=1).max(axis=1)).T

def Brank_ts(x):
    return x.rank()/x.rank().max()

def rolling_rank(data_array):
        return (data_array.argsort(axis = 0).argsort(axis = 0) + 1)[-1, :]
            
def ts_rank(data, timewindow = 5):
    data_array = data.to_numpy()
    ts_array = np.zeros(shape = data_array.shape)
    for raw_index in range(ts_array.shape[0] + 1):
        real_index = raw_index - 1
        if real_index < timewindow - 1:
            ts_array[real_index, :] = np.nan
        else:
            ts_array[real_index, :] = rolling_rank(data_array[real_index - (timewindow - 1): real_index + 1, :])
    return pd.DataFrame(ts_array, columns = data.columns,index = data.index)

def ts_min(data, timewindow = 3):
    data_array = data.to_numpy()
    ts_min = np.zeros (shape = data_array.shape)
    for raw_index in range(ts_min.shape[0] + 1):
        real_index = raw_index - 1
        if real_index < timewindow - 1:
            ts_min[real_index, :] = np.nan
        else:
            ts_min[real_index, :] = (data_array[real_index - (timewindow - 1): real_index + 1, :]).min(axis = 0)
    return pd.DataFrame(ts_min, columns = data.columns,index = data.index)

def ts_max(data, timewindow = 3):
    data_array = data.to_numpy()
    ts_max = np.zeros(shape = data_array.shape)
    for raw_index in range(ts_max.shape[0] + 1):
        real_index = raw_index - 1
        if real_index < timewindow - 1:
            ts_max[real_index, :] = np.nan
        else:
            ts_max[real_index, :] = (data_array[real_index - (timewindow - 1): real_index + 1, :]).max(axis = 0)
    return pd.DataFrame(ts_max, columns = data.columns,index = data.index)

def decay_linear(data, period = 3):
        data_array = data.to_numpy()
        na_lwma = np.zeros_like(data_array)
        na_lwma[:period, :] = data_array[:period, :]
        divisor = period * (period + 1) / 2
        y = (np.arange(period) + 1) * 1.0 / divisor
        for row in range(period - 1, data_array.shape[0]):
            x = data_array[row - period + 1: row + 1, :]
            na_lwma[row, :] = (np.dot(x.T, y))
        return pd.DataFrame(na_lwma, columns = data.columns,index = data.index)

def ts_argmax(data, timewindow = 3):
        data_array = data.to_numpy()
        ts_argmax = np.zeros(shape = data_array.shape)
        for raw_index in range(ts_argmax.shape[0] + 1):
            real_index = raw_index - 1
            if real_index < timewindow - 1:
                ts_argmax[real_index, :] = np.nan
            else:
                ts_argmax[real_index, :] = (data_array[real_index - (timewindow - 1): real_index + 1, :]).argmax(axis = 0)
        return pd.DataFrame(ts_argmax+1, columns = data.columns,index = data.index)

class Alpha101_1_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('1')

    def calc(self):
        spp = StockPostPriceFactor().get()
        ret = spp.apply(np.log).diff()
        std = ret.rolling(20).std()
        helper = ret[ret>0].apply(np.sign).replace(np.nan,0)
        x1 = (1 - helper) * std + helper * spp
        x2 = x1.apply(np.sign) * (x1 ** 2)
        x3 = x2.rolling(5).apply(lambda x: x.argmax(), raw=True)
        self._factor = Brank_cs(x3) - 0.5

class Alpha101_2_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('2')

    def calc(self):
        vlm = StockVolumeFactor().get()
        op = StockOpenPriceFactor().get()
        cl = StockPostPriceFactor().get()
        x1 = vlm.apply(np.log).diff(2)
        x2 = Brank_cs(x1)
        y1 = (cl-op)/op
        y2 = Brank_cs(y1)
        self._factor = -x2.rolling(6).corr(y2)

class Alpha101_3_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('3')

    def calc(self):
        vlm = StockVolumeFactor().get()
        op = StockOpenPriceFactor().get()
        x1 = Brank_cs(vlm)
        y1 = Brank_cs(op)
        self._factor = -x1.rolling(10).corr(y1)

class Alpha101_4_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('4')

    def calc(self):
        low = StockLowPriceFactor().get()
        x1 = Brank_cs(low)
        self._factor = ts_rank(x1)

class Alpha101_5_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('5')

    def calc(self):
        vwap =  StockVWAPFactor().get()
        op = StockOpenPriceFactor().get()
        cl = StockPostPriceFactor().get()
        x1 = Brank_cs(op - vwap.rolling(10).mean())
        y1 = -(Brank_cs((cl-vwap).apply(np.abs)))
        self._factor = x1*y1

class Alpha101_6_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('6')

    def calc(self):
        vlm = StockVolumeFactor().get()
        op = StockOpenPriceFactor().get()
        self._factor = op.rolling(10).corr(vlm)

class Alpha101_7_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('7')

    def calc(self):
        vlm = StockVolumeFactor().get()
        cl = StockPostPriceFactor().get()
        helper = vlm[vlm<vlm.rolling(20).mean()].apply(np.sign).replace(np.nan,0)
        x1 = helper * (-ts_rank(cl.diff(7).apply(np.abs))*cl.diff(7).apply(np.sign))
        self._factor = x1.replace(0,-1)

class Alpha101_8_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('8')

    def calc(self):
        cl = StockPostPriceFactor().get()
        op = StockOpenPriceFactor().get()
        ret = cl.apply(np.log).diff()
        x1 = op.rolling(5).sum()*ret.rolling(5).sum()
        self._factor = -Brank_cs(x1.diff(10))

class Alpha101_9_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('9')

    def calc(self):
        cl = StockPostPriceFactor().get()
        temp = ts_min(cl.diff(), timewindow=5)
        helper = temp[temp<0].apply(np.sign).replace(np.nan,0)
        temp2 = ts_max(cl.diff(), timewindow=4)
        helper2 = temp2[temp2<0].apply(np.sign).replace(np.nan,0)
        x1 = helper2 * cl.diff() + (1 - helper2) * -(cl.diff())        
        self._factor = helper * cl.diff() + (1 - helper) * x1

class Alpha101_10_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('10')

    def calc(self):
        cl = StockPostPriceFactor().get()
        op = StockOpenPriceFactor().get()
        ret = cl.apply(np.log).diff()
        x1 = op.rolling(5).sum()*ret.rolling(5).sum()
        self._factor = -Brank_cs(x1.diff(10))

class Alpha101_11_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('11')

    def calc(self):
        vwap =  StockVWAPFactor().get()
        cl = StockPostPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = Brank_cs(ts_max(vwap-cl))
        y1 = Brank_cs(ts_min(vwap-cl))
        z1 = Brank_cs(vlm.diff(3))
        self._factor = (x1 + y1) * z1

class Alpha101_12_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('12')

    def calc(self):
        cl = StockPostPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = vlm.diff().apply(np.sign)
        y1 = -cl.diff()
        self._factor = x1 * y1

class Alpha101_13_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('13')

    def calc(self):
        cl = StockPostPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = Brank_cs(cl)
        y1 = Brank_cs(vlm)
        z1 = x1.rolling(5).cov(y1)
        self._factor = -Brank_cs(z1)

class Alpha101_14_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('14')

    def calc(self):
        cl = StockPostPriceFactor().get()
        vlm = StockVolumeFactor().get()
        op = StockOpenPriceFactor().get()
        ret = cl.apply(np.log).diff()
        x1 = Brank_cs(ret.diff(3))
        y1 = op.rolling(10).corr(vlm)
        self._factor = -x1*y1

class Alpha101_15_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('15')

    def calc(self):
        high = StockHighPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = Brank_cs(high)
        x2 = Brank_cs(vlm)
        y1 = x1.rolling(3).corr(x2)
        y2 = Brank_cs(y1)
        self._factor = -y2.rolling(3).sum()

class Alpha101_16_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('16')

    def calc(self):
        high = StockHighPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = Brank_cs(high)
        y1 = Brank_cs(vlm)
        z1 = x1.rolling(5).cov(y1)
        self._factor = -Brank_cs(z1)

class Alpha101_17_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('17')

    def calc(self):
        cl = StockPostPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = -Brank_cs(ts_rank(cl, timewindow=10))
        y1 = Brank_cs((cl.diff()).diff())
        z1 = Brank_cs(ts_rank(vlm/vlm.rolling(20).mean(), 10))
        self._factor = x1*y1*z1

class Alpha101_18_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('18')

    def calc(self):
        cl = StockPostPriceFactor().get()
        op = StockOpenPriceFactor().get()
        x1 = ((cl-op).apply(np.abs)).rolling(5).std()+cl-op
        y1 = cl.rolling(10).corr(op)
        self._factor = -1*Brank_cs(x1+y1)

class Alpha101_19_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('19')

    def calc(self):
        cl = StockPostPriceFactor().get()
        ret = cl.apply(np.log).diff()
        x1 = -1*(cl-cl.shift(7)+cl.diff(7)).apply(np.sign)
        y1 = Brank_cs(1 + ret.rolling(250).sum())
        self._factor = x1 * y1

class Alpha101_20_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('20')

    def calc(self):
        cl = StockPostPriceFactor().get()
        op = StockOpenPriceFactor().get()
        high = StockHighPriceFactor().get()
        low = StockLowPriceFactor().get()
        x1 = -Brank_cs(op - high.shift(1))
        y1 = Brank_cs(op - cl.shift(1))
        z1 = Brank_cs(op - low.shift(1))
        self._factor = x1 * y1 * z1

class Alpha101_22_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('22')

    def calc(self):
        cl = StockPostPriceFactor().get()
        high = StockHighPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = (high.rolling(5).corr(vlm)).diff(5)
        y1 = Brank_cs(cl.rolling(20).std())
        self._factor = -1 * x1 * y1

class Alpha101_23_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('23')

    def calc(self):
        high = StockHighPriceFactor().get()
        helper = high[high.rolling(20).mean()<high].apply(np.sign).replace(np.nan,0)
        x1 = helper * -1 * (high.diff(2))
        self._factor = x1

class Alpha101_25_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('25')

    def calc(self):
        cl = StockPostPriceFactor().get()
        vwap =  StockVWAPFactor().get()
        high = StockHighPriceFactor().get()
        vlm = StockVolumeFactor().get()
        ret = cl.apply(np.log).diff()
        x1 = (-1 * ret) * vlm.rolling(20).mean() * vwap
        self._factor = Brank_cs(x1 * (high - cl))

class Alpha101_33_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('33')

    def calc(self):
        cl = StockPostPriceFactor().get()
        op = StockOpenPriceFactor().get()
        self._factor = Brank_cs(-1 * (1 - (op / cl)))

class Alpha101_34_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('34')

    def calc(self):
        cl = StockPostPriceFactor().get()
        ret = cl.apply(np.log).diff()
        x1 = Brank_cs(ret.rolling(2).std() / ret.rolling(5).std())
        x2 = Brank_cs(cl.diff())
        self._factor = (1 - x1) + (1 -x2)

class Alpha101_40_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('40')

    def calc(self):
        high = StockHighPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = Brank_cs(high.rolling(10).std())
        x2 = high.rolling(10).corr(vlm)
        self._factor = -1 * x1 * x2

class Alpha101_41_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('41')

    def calc(self):
        low = StockLowPriceFactor().get()
        vwap = StockVWAPFactor().get()
        high = StockHighPriceFactor().get()
        self._factor = ((low * high).apply(np.sqrt) - vwap)

class Alpha101_42_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('42')

    def calc(self):
        cl = StockPostPriceFactor().get()
        vwap = StockVWAPFactor().get()
        x1 = Brank_cs(vwap-cl)
        y1 = Brank_cs(vwap+cl)
        self._factor = x1 / y1

class Alpha101_43_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('43')

    def calc(self):
        cl = StockPostPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = ts_rank((vlm / vlm.rolling(20).mean()), 20)
        x2 = ts_rank((-1* cl.diff(7)), 8)
        self._factor = x1 * x2

class Alpha101_44_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('44')

    def calc(self):
        high = StockHighPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = Brank_cs(vlm)
        self._factor = -1 * high.rolling(5).corr(x1)


class Alpha101_53_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('53')

    def calc(self):
        cl = StockPostPriceFactor().get()
        low = StockLowPriceFactor().get() 
        high = StockHighPriceFactor().get()
        x1 = ((cl - low) - (high - cl))/(cl - low)
        self._factor = -1 * x1.diff(9)

class Alpha101_54_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('54')

    def calc(self):
        cl = StockPostPriceFactor().get()
        low = StockLowPriceFactor().get() 
        op = StockOpenPriceFactor().get()
        high = StockHighPriceFactor().get()
        x1 = (low - cl) * (op ** 5)/((low - high)*(cl ** 5))
        self._factor = -1 * x1

class Alpha101_55_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('55')

    def calc(self):
        cl = StockPostPriceFactor().get()
        low = StockLowPriceFactor().get() 
        high = StockHighPriceFactor().get()
        vlm = StockVolumeFactor().get()
        x1 = cl - ts_min(low, 12)
        x2 = ts_max(high, 12) - ts_min(low, 12)
        y1 = Brank_cs(x1 / x2)
        y2 = Brank_cs(vlm)
        self._factor = -1 * y1.rolling(6).corr(y2)

class Alpha101_56_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('56')

    def calc(self):
        cl = StockPostPriceFactor().get()
        cap = MarketValueFactor().get()
        ret = cl.apply(np.log).diff()
        x1 = ret.rolling(2).sum()
        x2 = x1.rolling(3).sum()
        x3 = ret.rolling(10).sum()
        y1 = x3/x2
        y2 = cap * ret
        self._factor = -1 * Brank_cs(y1) * Brank_cs(y2)

class Alpha101_57_Factor(Alpha101Factor):
    def __init__(self):
        super().__init__('57')

    def calc(self):
        cl = StockPostPriceFactor().get()
        vwap =  StockVWAPFactor().get()
        x1 = Brank_cs(ts_argmax(cl,30))
        self._factor = -1 * (cl - vwap)/decay_linear(x1, 2)



    

         















        

    







    










    



from typing import Dict

import pandas as pd

from .factor_types import GrowthFactor, MixFactorBase
from .basic_factors import IncomeFactor, AdjNetProfitFactor, CashFlowFactor
from .utils import calc_growth_rate, normalize


class IncomeGrowthFactor(GrowthFactor):
    # 收入增长
    def __init__(self):
        super().__init__('inc_gr')
        self._deps.add(IncomeFactor())

    def calc(self):
        # 过去四个季度的营业收入增长率
        income = IncomeFactor().get()
        self._factor = income.apply(calc_growth_rate, axis=1, result_type='broadcast', whole_df=income)


class NetProfitOfNonRecurringGoLFactor(GrowthFactor):
    # 扣非后盈利增长
    def __init__(self):
        super().__init__('np_nrgol')
        self._deps.add(AdjNetProfitFactor())

    def calc(self):
        # 过去四个季度扣除非经常性损益后的净利润的增长率
        adj_net_profit = AdjNetProfitFactor().get()
        self._factor = adj_net_profit.apply(calc_growth_rate, axis=1, result_type='broadcast', whole_df=adj_net_profit)


class OperationalCashFlowFactor(GrowthFactor):
    # 经营性现金流增长
    def __init__(self):
        super().__init__('op_cf')
        self._deps.add(CashFlowFactor())

    def calc(self):
        # 企业过去四个季度经营性现金流的增长率
        cash_flow = CashFlowFactor().get()
        self._factor = cash_flow.apply(calc_growth_rate, axis=1, result_type='broadcast', whole_df=cash_flow)


class MixGrowthFactor(MixFactorBase, GrowthFactor):
    # 合成因子
    def __init__(self):
        super(MixFactorBase, self).__init__('mixg')
        self._factor: Dict[pd.DataFrame] = {}
        self._deps.add(IncomeGrowthFactor())
        self._deps.add(NetProfitOfNonRecurringGoLFactor())
        self._deps.add(OperationalCashFlowFactor())

    def calc(self, universe: str):
        index = IncomeGrowthFactor().get(universe).index
        columns = IncomeGrowthFactor().get(universe).columns
        self._factor[universe] = pd.DataFrame((normalize(IncomeGrowthFactor().get(universe).T) + normalize(NetProfitOfNonRecurringGoLFactor().get(universe).T) +
                                               normalize(OperationalCashFlowFactor().get(universe).T)).T, index=index, columns=columns)
